import math
import os
from datetime import date

import requests
from prettytable import PrettyTable


class Docker:
    def __init__(self, options, auth):
        self.options = options
        self.auth = auth
        self.is_json = self.options.display == "JSON"

    def list(self):
        r = requests.get(os.path.join(self.options.url, "docker/list"),
                         auth=self.auth, params={'docker': self.options.docker})
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'docker/list\' service: %s' % r.text)
        result = r.json()
        if not self.is_json:
            res = PrettyTable(["Date", "IMAGE", "Tag", "Configurations"])
            res.align["Configurations"] = "l"
            for r in sorted(result, key=lambda r: float(r["date"])):
                d = date.fromtimestamp(math.ceil(float(r["date"] or 0))).isoformat()
                imgtag = r["image"].split(':')
                res.add_row([d, imgtag[0], imgtag[1], r["configs"]])
        else:
            res = result
        return res

    def add(self):
        data = {
            'image': self.options.image
        }

        if not os.path.exists(self.options.configs):
            raise RuntimeError('%s is not a file.' % self.options.configs)

        if not os.path.exists(self.options.schema):
            raise RuntimeError('%s is not a file.' % self.options.schema)

        with open(self.options.configs) as cfile:
            data['configs'] = cfile.read()

        with open(self.options.schema) as sfile:
            data['schema'] = sfile.read()

        r = requests.post(os.path.join(self.options.url, "docker/add"), auth=self.auth, data=data)

        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'docker/add\' service: %s' % r.text)

        res = 'ok'
        return res

    def describe(self):
        image = self.options.docker
        p = image.find(":")
        tag = image[p + 1:]
        image = image[:p]
        assert self.options.config, "docker describe requires --config parameter"
        r = requests.get(os.path.join(self.options.url, "docker/describe"),
                         params={'config': self.options.config, 'image': image, 'tag': tag},
                         auth=self.auth)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'service/describe\' service: %s' % r.text)
        res = r.json()
        return res

    def execute_command(self):
        result = None
        if self.options.subcmd == "list":
            result = self.list()
        if self.options.subcmd == "add":
            result = self.add()
        if self.options.subcmd == "describe":
            result = self.describe()
        return result



#!/usr/bin/env python
# coding: utf-8

import client.launcher as launcher


def define_option():
    launcher.subparsers_map["docker"] = launcher.subparsers.add_parser('docker')
    launcher.subparsers_map["model"] = launcher.subparsers.add_parser('model')
    launcher.subparsers_map["vocab"] = launcher.subparsers.add_parser('vocab')
    launcher.subparsers_map["resource"] = launcher.subparsers.add_parser('resource')
    launcher.subparsers_map["user"] = launcher.subparsers.add_parser('user')
    launcher.subparsers_map["tag"] = launcher.subparsers.add_parser('tag')
    launcher.subparsers_map["auth"] = launcher.subparsers.add_parser('auth')

    launcher.subparsers_map["entity"] = launcher.subparsers.add_parser('entity')
    launcher.subparsers_map["permission"] = launcher.subparsers.add_parser('permission')
    launcher.subparsers_map["role"] = launcher.subparsers.add_parser('role')
    launcher.subparsers_map["group"] = launcher.subparsers.add_parser('group')
    launcher.subparsers_map["share"] = launcher.subparsers.add_parser('share')
    launcher.subparsers_map["monitoring"] = launcher.subparsers.add_parser('monitoring')

    subparsers_auth = launcher.subparsers_map["auth"].add_subparsers(help='sub-command help',
                                                                     dest='subcmd')
    subparsers_auth.required = True

    parser_login = subparsers_auth.add_parser('login',
                                              help='{login} login to systran-nmt-wizard')
    launcher.shortcut_map["login"] = ["auth", "login"]
    parser_login.add_argument('-u', '--user', help='user')

    parser_auth = subparsers_auth.add_parser('token', help='{auth} get auth token for other user '
                                                           '(super only)')
    launcher.shortcut_map["auth"] = ["auth", "token"]
    parser_auth.add_argument('-u', '--user', required=True, help='user')
    parser_auth.add_argument('--duration', type=int, help='specify duration of token ')
    parser_auth.add_argument('--persistent', action='store_true',
                             help='use persistent for generating persistent api keys')

    parser_revoke = subparsers_auth.add_parser('revoke',
                                               help='{revoke} revoke a token (super only)')
    launcher.shortcut_map["revoke"] = ["auth", "revoke"]
    parser_revoke.add_argument('-T', '--token', type=str, required=True, help='the token to revoke')

    parser_logout = subparsers_auth.add_parser('logout',
                                               help='{logout} revoke current credentials')
    launcher.shortcut_map["logout"] = ["auth", "logout"]

    subparsers_vocab = launcher.subparsers_map["vocab"].add_subparsers(help='sub-command help',
                                                                       dest='subcmd')
    subparsers_vocab.required = True

    parser_get_vocab = subparsers_vocab.add_parser('get', help='download the vocab')
    parser_get_vocab.add_argument('vocab', help='vocab')
    parser_get_vocab.add_argument('-f', '--file', help='path to the file of the vocab')
    parser_get_vocab.add_argument('-o', '--output', help='path to the output file of the vocab')

    parser_detail_vocab = subparsers_vocab.add_parser('detail', help='{cv} show the model contents')
    launcher.shortcut_map["cv"] = ["vocab", "detail"]
    parser_detail_vocab.add_argument('vocab')
    parser_detail_vocab.add_argument('-d', '--directory', help='the directory inside the vocab')

    subparsers_model = launcher.subparsers_map["model"].add_subparsers(help='sub-command help',
                                                                       dest='subcmd')
    subparsers_model.required = True

    parser_list_models = subparsers_model.add_parser('list',
                                                     help='{lm} list available models')
    launcher.shortcut_map["lm"] = ["model", "list"]
    parser_list_models.add_argument('-s', '--source', help='source language prefix')
    parser_list_models.add_argument('-t', '--target', help='target language prefix')
    parser_list_models.add_argument('-m', '--model', help='beginning pattern on model name')
    parser_list_models.add_argument('--skip_noscores', action='store_true',
                                                     help='skip models without scores')
    parser_list_models.add_argument('--has_noscores', action='store_true',
                                    help='only display models with scores missing')
    parser_list_models.add_argument('--quiet', '-q', action='store_true',
                                    help='only display matching model ids')
    parser_list_models.add_argument('--show_owner', '-so', action='store_true',
                                    help='display model owner')
    parser_list_models.add_argument('--count', action='store_true',
                                    help='only display count of models')
    parser_list_models.add_argument('--aggr', choices=['lp', 'model'],
                                    help='aggregate models by `lp` or `model`')
    parser_list_models.add_argument('--scores', nargs='*', default=None,
                                    help='testset patterns to display along with the model')
    parser_list_models.add_argument('--metric', '-M', default='BLEU',
                                    help='metric to display (default BLEU)')
    parser_list_models.add_argument('--show_pruned', action='store_true',
                                    help='display pruned and not pruned models')
    parser_list_models.add_argument('--only_show_pruned', action='store_true',
                                    help='only display pruned models')

    parser_get_model = subparsers_model.add_parser('get', help='download the model')
    parser_get_model.add_argument('model', help='model')
    parser_get_model.add_argument('-f', '--file', help='path to the file of the model')
    parser_get_model.add_argument('-o', '--output', help='path to the output file of the model')

    parser_detail_model = subparsers_model.add_parser('detail', help='{cm} show the model contents')
    launcher.shortcut_map["cm"] = ["model", "detail"]
    parser_detail_model.add_argument('model')
    parser_detail_model.add_argument('-d', '--directory', help='the directory inside the model')

    parser_delete_models = subparsers_model.add_parser('delete',
                                                       help='{dm} delete specific models')
    launcher.shortcut_map["dm"] = ["model", "delete"]
    parser_delete_models.add_argument('-s', '--source', help='source language', required=True)
    parser_delete_models.add_argument('-t', '--target', help='target language', required=True)
    parser_delete_models.add_argument('--recursive', action='store_true',
                                      help='recursive deletion of each model and its descendant')
    parser_delete_models.add_argument('-f', '--force', action='store_true',
                                      help='do not ask confirmation of the deletion')
    parser_delete_models.add_argument('-d', '--dryrun', action='store_true',
                                      help='just simulate deletion to show models impacted')
    parser_delete_models.add_argument('models', nargs='+', type=str, help='model names')

    parser_add_model = subparsers_model.add_parser('add', help='{am} upload a model')
    launcher.shortcut_map["am"] = ["model", "add"]
    parser_add_model.add_argument('-f', '--file', help='path to tgz archive containing the model',
                                  required=True)
    parser_add_model.add_argument('--ignore_parent', action='store_true',
                                  help='if model parent not in catalog, drop connection '
                                       'to parent_model')
    parser_add_model.add_argument('--compute_checksum', action='store_true',
                                  help='recompute checksum during upload process')
    parser_add_model.add_argument('--name', '-N', type=str,
                                  help='rename during upload process')

    parser_describe_model = subparsers_model.add_parser('describe',
                                                        help='describe a model')
    parser_describe_model.add_argument('model', help='model')

    parser_addtag_model = subparsers_model.add_parser('tagadd', help='add tag[s] to a model')
    parser_addtag_model.add_argument('-m', '--model', help='model', required=True)
    parser_addtag_model.add_argument('-t', '--tags', nargs='+', type=str, help='tag names',
                                     required=True)
    parser_adddel_model = subparsers_model.add_parser('tagdel', help='remove tag[s] from a model')
    parser_adddel_model.add_argument('-m', '--model', help='model', required=True)
    parser_adddel_model.add_argument('-t', '--tags', nargs='+', type=str, help='tag names',
                                     required=True)

    parser_share_model = subparsers_model.add_parser('share', help='share a model with an entity')
    launcher.shortcut_map["ms"] = ["model", "share"]
    parser_share_model.add_argument('-m', '--model', help='model', required=True)
    parser_share_model.add_argument('-e', '--entity_code', help='2-letters entity code', required=True)

    parser_removeshare_model = subparsers_model.add_parser('removeshare',
                                                           help='remove the share visibility '
                                                                'of a model for an entity')
    launcher.shortcut_map["mrs"] = ["model", "removeshare"]
    parser_removeshare_model.add_argument('-m', '--model', help='model', required=True)
    parser_removeshare_model.add_argument('-e', '--entity_code', help='2-letters entity code', required=True)

    parser_open_model = subparsers_model.add_parser('open', help='open a model with an entity')
    launcher.shortcut_map["mo"] = ["model", "open"]
    parser_open_model.add_argument('-m', '--model', help='model', required=True)
    parser_open_model.add_argument('-e', '--entity_code', help='2-letters entity code', required=True)

    parser_removeopen_model = subparsers_model.add_parser('removeopen',
                                                           help='remove the open visibility '
                                                                'of a model for an entity')
    launcher.shortcut_map["mro"] = ["model", "removeopen"]
    parser_removeopen_model.add_argument('-m', '--model', help='model', required=True)
    parser_removeopen_model.add_argument('-e', '--entity_code', help='2-letters entity code', required=True)

    subparsers_docker = launcher.subparsers_map["docker"].add_subparsers(help='sub-command help',
                                                                         dest='subcmd')
    subparsers_docker.required = True

    parser_list_dockers = subparsers_docker.add_parser('list',
                                                       help='{ld} list available dockers')
    launcher.shortcut_map["ld"] = ["docker", "list"]
    parser_list_dockers.add_argument('-d', '--docker', default="",
                                     help='restrict to specific docker')

    parser_add_docker = subparsers_docker.add_parser('add', help='{ad} add docker configs and '
                                                                 'schema to db')
    launcher.shortcut_map["ad"] = ["docker", "add"]
    parser_add_docker.add_argument('-i', '--image', required=True,
                                   help='image full name')
    parser_add_docker.add_argument('-c', '--configs', required=True,
                                   help='docker configs')
    parser_add_docker.add_argument('-s', '--schema', required=True,
                                   help='schema for docker configs')

    parser_describe_docker = subparsers_docker.add_parser('describe',
                                                          help='describe a docker image')
    parser_describe_docker.add_argument('docker', help='docker to describe')
    parser_describe_docker.add_argument('-c', '--config', help='name of the config', required=True)

    subparsers_resource = launcher.subparsers_map["resource"].add_subparsers(help='sub-command '
                                                                                  'help',
                                                                             dest='subcmd')
    subparsers_resource.required = True

    parser_list_resources = subparsers_resource.add_parser('list',
                                                           help='{lr} list available resources')
    launcher.shortcut_map["lr"] = ["resource", "list"]
    parser_list_resources.add_argument('-s', '--service', help="service name")
    parser_list_resources.add_argument('path', nargs='?', default=None, help='subpath')
    parser_list_resources.add_argument('--aggr', action='store_true',
                                       help='aggregate file by suffixes')

    parser_get_resources = subparsers_resource.add_parser('get', help='{gr} download a file from '
                                                                      'given storage')
    launcher.shortcut_map["gr"] = ["resource", "get"]
    parser_get_resources.add_argument('-s', '--service', help="service name")
    parser_get_resources.add_argument('--path', help='path to the file', required=True)

    subparsers_user = launcher.subparsers_map["user"].add_subparsers(help='sub-command help',
                                                                     dest='subcmd')
    subparsers_user.required = True
    parser_list_users = subparsers_user.add_parser('list',
                                                   help='{lu} list users')
    launcher.shortcut_map["lu"] = ["user", "list"]

    parser_who = subparsers_user.add_parser('whoami', help='{who} show current user info')
    launcher.shortcut_map["who"] = ["user", "whoami"]

    parser_add_user = subparsers_user.add_parser('add', help='{au} add user')
    launcher.shortcut_map["au"] = ["user", "add"]
    parser_add_user.add_argument('--first_name', help='first name', required=True)
    parser_add_user.add_argument('--last_name', help='last name', required=True)
    parser_add_user.add_argument('-e', '--email', help='user email', required=True)
    parser_add_user.add_argument('-p', '--password', help='user password', required=True)
    parser_add_user.add_argument('--user_code', help='3-letter user code', required=True)
    parser_add_user.add_argument('--entity', help='entity id/name/code', required=True)
    parser_add_user.add_argument('--roles', nargs='*', help='roles')
    parser_add_user.add_argument('--groups', nargs='*', help='groups')

    parser_mod_user = subparsers_user.add_parser('modify', help='{mu} change user parameters')
    launcher.shortcut_map["mu"] = ["user", "modify"]
    parser_mod_user.add_argument('-u', '--user', help='user id/email/code', required=True)
    parser_mod_user.add_argument('--first_name', help='first name')
    parser_mod_user.add_argument('--last_name', help='last name')
    parser_mod_user.add_argument('-p', '--password', help='password')
    parser_mod_user.add_argument('--roles', nargs='*', help='roles')
    parser_mod_user.add_argument('--groups', nargs='*', help='groups')

    parser_del_user = subparsers_user.add_parser('delete', help='{du} remove user')
    launcher.shortcut_map["du"] = ["user", "delete"]
    parser_del_user.add_argument('-u', '--user', help='user id/email/code', required=True)

    parser_pwd_user = subparsers_user.add_parser('password', help='{password} change password')
    launcher.shortcut_map["password"] = ["user", "password"]
    parser_pwd_user.add_argument('-p', '--password', help='password', required=True)

    subparsers_tag = launcher.subparsers_map["tag"].add_subparsers(help='sub-command help',
                                                                   dest='subcmd')
    subparsers_tag.required = True
    parser_list_tags = subparsers_tag.add_parser('list', help='list tags')
    parser_add_tag = subparsers_tag.add_parser('add', help='add tag')
    parser_add_tag.add_argument('name', help='tag name')
    parser_detail_tag = subparsers_tag.add_parser('detail', help='show tag detail')
    parser_detail_tag.add_argument('name', help='tag name')
    parser_del_tag = subparsers_tag.add_parser('delete', help='remove tag')
    parser_del_tag.add_argument('name', help='tag name')
    parser_modify_tag = subparsers_tag.add_parser('modify', help='update tag name')
    parser_modify_tag.add_argument('name', help='current tag name')
    parser_modify_tag.add_argument('-n', '--newname', help='new tag name', required=True)

    parser_change_tasks = launcher.subparsers_tasks.add_parser('change',
                                                               help='{ct} change queued task')
    launcher.shortcut_map["ct"] = ["task", "change"]
    parser_change_tasks.add_argument('-p', '--prefix',
                                     help='prefix for the tasks to change')
    parser_change_tasks.add_argument('-P', '--priority', type=int,
                                     help='task priority - highest better')
    parser_change_tasks.add_argument('-Pr', '--priority_rand', type=int, default=0,
                                     help='for each task add this random number to priority')
    parser_change_tasks.add_argument('-s', '--service',
                                     help="service name")
    parser_change_tasks.add_argument('-g', '--gpus',
                                     help="number of gpus", type=int)
    parser_change_tasks.add_argument('task_ids', nargs='*',
                                     help="task identifiers")

    launcher.parser_launch.add_argument('-N', '--no_test_trans', action='store_true',
                                        help="disable automatic test file translations")
    launcher.parser_launch.add_argument('--novalidschema', '--no_valid_schema', action='store_true',
                                        help='skip config validation')
    launcher.parser_launch.add_argument('--upgrade', choices=['auto', 'none', 'force'],
                                        default='auto',
                                        help='choice to upgrade when later docker image available:'
                                             ' `auto`(interactive), `none`, `force`')

    parser_config_service = launcher.subparsers_service.add_parser('config',
                                                                   help='configure a service')
    parser_config_service.add_argument('action',
                                       help="command list, set, get, del, select")
    parser_config_service.add_argument('-s', '--service', required=True,
                                       help="service name")
    parser_config_service.add_argument('-cn', '--configname',
                                       help="configuration name")
    parser_config_service.add_argument('-c', '--config',
                                       help="configuration file (for `set` only)")
    launcher.shortcut_map["listconfig"] = ["service", "config", "list"]
    launcher.shortcut_map["setconfig"] = ["service", "config", "set"]
    launcher.shortcut_map["getconfig"] = ["service", "config", "get"]
    launcher.shortcut_map["delconfig"] = ["service", "config", "del"]
    launcher.shortcut_map["selectconfig"] = ["service", "config", "select"]

    parser_disable_service = launcher.subparsers_service.add_parser('disable',
                                                                    help='disable a resource')
    parser_disable_service.add_argument('-r', '--resource', required=True,
                                        help="name of the resource")
    parser_disable_service.add_argument('-s', '--service', required=True,
                                        help="service name")
    parser_disable_service.add_argument('-m', '--message', required=True,
                                        help="add disabling message")

    parser_enable_service = launcher.subparsers_service.add_parser('enable',
                                                                   help='enable a resource')
    parser_enable_service.add_argument('-s', '--service', required=True,
                                       help="service name")
    parser_enable_service.add_argument('-r', '--resource', required=True,
                                       help="name of the resource")

    parser_stop_service = launcher.subparsers_service.add_parser('stop', help='stop a pool worker')
    parser_stop_service.add_argument('-s', '--service', help="service name", required=True)

    parser_restart_service = launcher.subparsers_service.add_parser('restart',
                                                                    help='restart a pool worker')
    parser_restart_service.add_argument('-s', '--service', help="service name", required=True)

    subparsers_entity = launcher.subparsers_map["entity"].add_subparsers(help='sub-command help',
                                                                         dest='subcmd')
    subparsers_entity.required = True
    parser_list_entities = subparsers_entity.add_parser('list',
                                                        help='list entities')
    parser_list_entities.add_argument('-a', '--all', action='store_true',
                                      help='show all entity (include disabled entities)')

    parser_add_entity = subparsers_entity.add_parser('add',
                                                     help='add a new entity')
    parser_add_entity.add_argument('--entity_code', help='2-letter entity code', required=True)
    parser_add_entity.add_argument('--name', help='entity name', required=True)
    parser_add_entity.add_argument('-e', '--email', help='email', required=True)
    parser_add_entity.add_argument('--tel', help='contact telephone')
    parser_add_entity.add_argument('--address', help='address of entity')
    parser_add_entity.add_argument('--description', help='description of entity')
    parser_modify_entity = subparsers_entity.add_parser('modify',
                                                        help='modify an entity')
    parser_modify_entity.add_argument('--entity', help='entity name/id/code', required=True)
    parser_modify_entity.add_argument('--name', help='new entity name')
    parser_modify_entity.add_argument('-e', '--email', help='email')
    parser_modify_entity.add_argument('--tel', help='entity telephone')
    parser_modify_entity.add_argument('--address', help='address of entity')
    parser_modify_entity.add_argument('--description', help='description of entity')

    # TODO implement it on server side
    # parser_delete_entity = subparsers_entity.add_parser('delete', help='delete an entity')
    # parser_delete_entity.add_argument('--entity', help='entity name/id/code', required=True)

    parser_disable_entity = subparsers_entity.add_parser('disable', help='disable an entity')
    parser_disable_entity.add_argument('--entity', help='entity name/id/code', required=True)

    parser_enable_entity = subparsers_entity.add_parser('enable', help='enable an entity')
    parser_enable_entity.add_argument('--entity', help='entity name/id/code', required=True)

    subparsers_role = launcher.subparsers_map["role"].add_subparsers(help='sub-command help',
                                                                     dest='subcmd')
    subparsers_role.required = True
    parser_list_roles = subparsers_role.add_parser('list',
                                                   help='list roles')
    parser_add_role = subparsers_role.add_parser('add',
                                                 help='add a new role for an entity')
    parser_add_role.add_argument('--name', help='role name', required=True)
    parser_add_role.add_argument('--entity', help='entity id/code/name', required=True)
    parser_add_role.add_argument('--permissions', help='permissions (possible prefix by '
                                                       'entity code) associated to the role',
                                 nargs='+', type=str)
    parser_modify_role = subparsers_role.add_parser('modify',
                                                    help='modify a role for an entity')
    parser_modify_role.add_argument('--role', help='role name or id', required=True)
    parser_modify_role.add_argument('--name', help='role new name')
    parser_modify_role.add_argument('--permissions', help='permissions (possible prefix by '
                                                          'entity code) associated to the role',
                                    nargs='*', type=str)
    parser_delete_role = subparsers_role.add_parser('delete',
                                                    help='modify a role for an entity')
    parser_delete_role.add_argument('--role', help='role name or id', required=True)

    subparsers_share = launcher.subparsers_map["share"].add_subparsers(help='sub-command help',
                                                                       dest='subcmd')
    parser_add_share = subparsers_share.add_parser('add',
                                                   help='share a role from an entity to '
                                                        'another entity')
    parser_add_share.add_argument('--roles', help='role name or id', nargs="+", required=True)
    parser_add_share.add_argument('--src_entity', help='source entity', required=True)
    parser_add_share.add_argument('--dest_entity', help='destination entity', required=True)

    parser_delete_share = subparsers_share.add_parser('delete',
                                                      help='remove a shared role from an entity to '
                                                           'another entity')
    parser_delete_share.add_argument('--roles', help='role name or id', nargs="+", required=True)
    parser_delete_share.add_argument('--src_entity', help='source entity', required=True)
    parser_delete_share.add_argument('--dest_entity', help='destination entity', required=True)

    subparsers_group = launcher.subparsers_map["group"].add_subparsers(help='sub-command help',
                                                                       dest='subcmd')
    subparsers_group.required = True
    parser_list_groups = subparsers_group.add_parser('list',
                                                     help='list groups')
    parser_add_group = subparsers_group.add_parser('add',
                                                   help='add a new group for an entity')
    parser_add_group.add_argument('--name', help='group name', required=True)
    parser_add_group.add_argument('--entity', type=str, help='entity id/code/name')
    parser_add_group.add_argument('--users', nargs='*', help='users')
    parser_add_group.add_argument('--roles', nargs='*', help='roles')

    parser_modify_group = subparsers_group.add_parser('modify',
                                                      help='modify a group for an entity')
    parser_modify_group.add_argument('--group', help='group name or id', required=True)
    parser_modify_group.add_argument('--name', help='group new name')
    parser_modify_group.add_argument('--users', nargs='*', help='users')
    parser_modify_group.add_argument('--roles', nargs='*', help='roles')

    parser_delete_group = subparsers_group.add_parser('delete',
                                                      help='modify a group for an entity')
    parser_delete_group.add_argument('--group', help='group name or id', required=True)

    subparsers_permission = launcher.subparsers_map["permission"].add_subparsers(help='sub-command '
                                                                                      'help',
                                                                                 dest='subcmd')
    subparsers_permission.required = True
    parser_list_permissions = subparsers_permission.add_parser('list',
                                                               help='list permissions')



import logging
import os

import requests
from prettytable import PrettyTable

import client.launcher as launcher

launcher.LOGGER = logging.getLogger()


class Tag:
    def __init__(self, options, auth):
        self.options = options
        self.auth = auth
        self.is_json = self.options.display == "JSON"

    def list(self):
        r = requests.get(os.path.join(self.options.url, "model/tags"), auth=self.auth)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'tag/list\' service: %s' % r.text)
        result = r.json()
        if not self.is_json:
            res = PrettyTable(["Name", "Entity"])
            res.align["Name"] = "l"
            for r in result:
                res.add_row([r["tag"], r["entity"]])
        else:
            res = result
        return res

    def delete(self):
        r = requests.delete(os.path.join(self.options.url, "model/tag"), auth=self.auth,
                            params={'tag': self.options.name})
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'tag/delete\' service: %s' % r.text)
        res = 'ok delete tag %s' % self.options.name
        return res
    
    def add(self):
        r = requests.put(os.path.join(self.options.url, "model/tag"), auth=self.auth,
                         params={'tag': self.options.name})
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'tag/add\' service: %s' % r.text)
        tag = r.json()
        res = "ok new tag: name=%s, entity=%s, creator=%s" % (
            tag['tag'], tag['entity'], tag['creator'])
        return res
    
    def detail(self):
        r = requests.get(os.path.join(self.options.url, "model/tags", self.options.name),
                         auth=self.auth)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'tag/detail\' service: %s' % r.text)
        result = r.json()
        if not self.is_json:
            res = PrettyTable(["Name", "Entity", "Creator", "Models"])
            res.align["Name"] = "l"
            model_name = ""
            if "models" in result:
                model_name = ",".join(result["models"])
            res.add_row([result["tag"], result["entity"], result["creator"], model_name])
        else:
            res = result
        return res
    
    def modify(self):
        r = requests.post(os.path.join(self.options.url, "model/tag"), auth=self.auth,
                          params={'tag': self.options.name, 'newTag': self.options.newname})
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'tag/modify\' service: %s' % r.text)
        res = 'ok modify tag %s to %s' % (self.options.name, self.options.newname)
        return res

    def execute_command(self):
        result = None
        if self.options.subcmd == "add":
            result = self.add()
        if self.options.subcmd == "list":
            result = self.list()
        if self.options.subcmd == "detail":
            result = self.detail()
        if self.options.subcmd == "delete":
            result = self.delete()   
        if self.options.subcmd == "modify":
            result = self.modify()
        return result



import gzip
import json
import logging
import os
import sys
from collections import Counter
import tarfile

import requests
import six
from io import BytesIO
from prettytable import PrettyTable

from lib import utils
import client.launcher as launcher

launcher.LOGGER = logging.getLogger()


class Model:
    def __init__(self, options, auth):
        self.options = options
        self.auth = auth
        self.is_json = self.options.display == "JSON"

    def list(self):
        if self.options.skip_noscores and self.options.scores is None:
            raise RuntimeError('cannot use --skip_noscores without --scores')
        if self.options.has_noscores and self.options.scores is None:
            raise RuntimeError('cannot use --has_noscores without --scores')
        if self.options.has_noscores and self.options.skip_noscores:
            raise RuntimeError('cannot use --has_noscores with --skip_noscores')
        if self.options.show_pruned and self.options.only_show_pruned:
            raise RuntimeError('cannot use --show_pruned with --only_show_pruned')
        params = {'source': self.options.source,
                  'target': self.options.target,
                  'model': self.options.model}
        if self.options.scores is not None:
            params['scores'] = ",".join(self.options.scores)
        if self.options.count:
            r = requests.get(os.path.join(self.options.url, "model/lp/list"), auth=self.auth)
            if r.status_code != 200:
                raise RuntimeError('incorrect result from \'model/lp/list\' service: %s' % r.text)
            response = r.json()
            if not self.is_json:
                res = PrettyTable(["LP", "#Models"])
                for item in response:
                    res.add_row([item["lp"], int(item["count_model"])])
            else:
                res = response
        else:
            r = requests.get(os.path.join(self.options.url, "model/list"),
                             params=params, auth=self.auth)
            if r.status_code != 200:
                raise RuntimeError('incorrect result from \'model/list\' service: %s' % r.text)
            response = r.json()
            result = []
            metrics = Counter()
            for item in response:
                if self.options.model and self.options.model not in item['model']:
                    continue
                if self.options.show_pruned or (self.options.show_pruned is False and
                                                self.options.only_show_pruned is False and item["pruned"] is False) or\
                        (self.options.only_show_pruned and item["pruned"] is True):
                    if self.options.scores is not None:
                        new_scores = {}
                        for p, v in six.iteritems(item['scores']):
                            if not self.is_json:
                                if isinstance(v, float):
                                    v = {'BLEU': v}
                                elif isinstance(v, dict) and "score" in v:
                                    v = v['score']
                                for m in v:
                                    metrics[m] += 1
                                v = v.get(self.options.metric)
                            if v is not None:
                                new_scores[p] = v
                        item['scores'] = new_scores
                    result.append(item)
            if not self.is_json:
                scorenames = {}
                bestscores = {}

                # Calculate the aggregate sentence feed
                idx_result = {}
                root = []
                for r in result:
                    r['children_models'] = []
                    idx_result[r['lp'] + ":" + r['model']] = r
                for k, v in six.iteritems(idx_result):
                    parent_model = v['parent_model']
                    if 'parent_model' in v and v['parent_model'] is not None and \
                            v['lp'] + ":" + v['parent_model'] in idx_result:
                        p = v['lp'] + ":" + v['parent_model']
                        idx_result[p]['children_models'].append(k)
                    else:
                        root.append(k)
                utils.cum_sentenceCount(root, idx_result, 0)

                idx_result = {}
                root = []
                if self.options.aggr:
                    aggr_result = {}
                    for r in result:
                        model = r["model"]
                        q = model.find("_")
                        if q != -1:
                            q = model.find("_", q + 1)
                            model = model[q + 1:]
                            q = model.find("_")
                            if q != -1:
                                model = model[:q]
                        lpmodel = r["lp"]
                        if self.options.aggr == 'model':
                            lpmodel += ":" + model
                        if lpmodel not in aggr_result:
                            line_data = {'lp': r["lp"], 'cumSentenceCount': 0, 'date': 0,
                                         'model': '', 'scores': {}, 'count': 0,
                                         'imageTag': ''}
                            if self.options.show_owner:
                                line_data['owner'] = ''

                            aggr_result[lpmodel] = line_data
                            if self.options.aggr == 'model':
                                aggr_result[lpmodel]["imageTag"] = r["imageTag"]
                                aggr_result[lpmodel]["model"] = model
                                if self.options.show_owner:
                                    owner_obj = r.get('owner')
                                    aggr_result[lpmodel]["owner"] = owner_obj[
                                        "entity_code"] if owner_obj else ""

                        aggr_result[lpmodel]['count'] += 1
                        for s, v in six.iteritems(r['scores']):
                            if s not in aggr_result[lpmodel]['scores'] or \
                                    aggr_result[lpmodel]['scores'][s] < v:
                                aggr_result[lpmodel]['scores'][s] = v
                        if r["date"] > aggr_result[lpmodel]['date']:
                            aggr_result[lpmodel]['date'] = r["date"]
                        if r["cumSentenceCount"] > aggr_result[lpmodel]['cumSentenceCount']:
                            aggr_result[lpmodel]['cumSentenceCount'] = r["cumSentenceCount"]
                    result = [aggr_result[k] for k in aggr_result]
                for r in result:
                    r['children_models'] = []
                    lpmodel = r["lp"] + ":" + r["model"]
                    if 'parent_model' in r and r['parent_model'] is not None:
                        r["parent_model"] = r["lp"] + ':' + r["parent_model"]
                    idx_result[lpmodel] = r
                    for s, v in six.iteritems(r['scores']):
                        scorenames[s] = scorenames.get(s, 0) + 1
                        if s not in bestscores or v > bestscores[s]:
                            bestscores[s] = v
                for k, v in six.iteritems(idx_result):
                    if 'parent_model' in v and v['parent_model'] in idx_result:
                        p = v['parent_model']
                        idx_result[p]['children_models'].append(k)
                    else:
                        root.append(k)
                max_depth = utils.tree_depth(0, root, idx_result)
                model_maxsize = max_depth + 42
                scorenames_key = sorted(scorenames.keys())
                scoretable = []
                scorecols = []
                for i in range(len(scorenames_key)):
                    scorecols.append("T%d" % (i + 1))
                    scoretable.append("\tT%d:\t%s\t%d" % (i + 1, scorenames_key[i],
                                                          scorenames[scorenames_key[i]]))
                if self.options.quiet:
                    res = []
                    utils.tree_display(res, 0, root, idx_result, model_maxsize,
                                       scorenames_key, bestscores, self.options.skip_noscores,
                                       self.options.has_noscores, self.options.show_owner,
                                       self.options.show_pruned, self.options.only_show_pruned,
                                       self.options.quiet)
                else:
                    header = ["Date", "LP", "Type", "Model ID", "#Sentences"]
                    if self.options.show_owner:
                        header.append("Owner")
                    if self.options.show_pruned:
                        header.append("Pruned")
                    res1 = PrettyTable(header + scorecols)
                    res1.align["Model ID"] = "l"
                    utils.tree_display(res1, 0, root, idx_result, model_maxsize,
                                       scorenames_key, bestscores, self.options.skip_noscores,
                                       self.options.has_noscores, self.options.show_owner,
                                       self.options.show_pruned, self.options.only_show_pruned,
                                       self.options.quiet)
                    res = [res1]
                    res.append('* TOTAL: %d models\n' % len(result))
                    if metrics:
                        res.append("* AVAILABLE METRICS: %s" % ", ".join(metrics.keys()))
                    if len(scoretable):
                        res.append("* TESTSET:")
                        res.append('\n'.join(scoretable) + "\n")
            else:
                res = result
        return res

    def describe(self):
        r = requests.get(os.path.join(self.options.url, "model/describe", self.options.model),
                         auth=self.auth)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'service/describe\' service: %s' % r.text)
        res = r.json()
        return res

    def tagadd(self):
        taglist = []
        for tag in self.options.tags:
            taglist.append({'tag': tag})
        data = {
            'tags': taglist
        }
        r = requests.put(os.path.join(self.options.url, "model", self.options.model, "tags"),
                         auth=self.auth, json=data)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/tagadd\' service: %s' % r.text)
        res = "ok new tags \"%s\" attached to model %s" % (",".join(self.options.tags),
                                                           self.options.model)
        return res
    
    def tagdel(self):
        taglist = []
        for tag in self.options.tags:
            taglist.append({'tag': tag})
        data = {
            'tags': taglist
        }
        r = requests.delete(os.path.join(self.options.url, "model", self.options.model, "tags"),
                            auth=self.auth,
                            json=data)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/tagadd\' service: %s' % r.text)
        res = "ok tags \"%s\" are removed from model %s" % (",".join(self.options.tags),
                                                            self.options.model)
        return res
    
    def share(self):
        data = {
            'visibility': 'share',
            'model': self.options.model,
            'entity': self.options.entity_code
        }
        r = requests.post(os.path.join(self.options.url, "model", "visibility", "add"),
                          auth=self.auth, json=data)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/share\' service: %s' % r.text)
        res = 'ok model %s shared with entity %s' % (self.options.model, self.options.entity_code)
        return res
    
    def removeshare(self):
        data = {
            'visibility': 'share',
            'model': self.options.model,
            'entity': self.options.entity_code
        }
        r = requests.post(os.path.join(self.options.url, "model", "visibility", "delete"),
                          auth=self.auth,
                          json=data)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/removeshare\' service: %s' % r.text)
        res = 'ok share visibility removed on model %s for entity %s' % \
              (self.options.model, self.options.entity_code)
        return res
    
    def open(self):
        data = {
            'visibility': 'open',
            'model': self.options.model,
            'entity': self.options.entity_code
        }
        r = requests.post(os.path.join(self.options.url, "model", "visibility", "add"),
                          auth=self.auth,
                          json=data)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/open\' service: %s' % r.text)
        res = 'ok model %s opened with entity %s' % (self.options.model, self.options.entity_code)
        return res
    
    def removeopen(self):
        data = {
            'visibility': 'open',
            'model': self.options.model,
            'entity': self.options.entity_code
        }
        r = requests.post(os.path.join(self.options.url, "model", "visibility", "delete"),
                          auth=self.auth,
                          json=data)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/removeopen\' service: %s' % r.text)
        res = 'ok open visibility removed on model %s for entity %s' % \
              (self.options.model, self.options.entity_code)
        return res
    
    def get(self):
        with open(os.path.expanduser(self.options.output), 'w+') \
                if self.options.output else sys.stdout as output:
            r = requests.get(os.path.join(self.options.url, "model/getfile/",
                                          self.options.model, self.options.file),
                             params={'is_compressed': True}, auth=self.auth)
            if r.status_code != 200:
                raise RuntimeError('incorrect result from \'model/getfile\' service: %s' % r.text)

            if self.options.output:
                output.write(gzip.GzipFile('', 'r', 0, BytesIO(r.content)).read())
            else:
                utils.write_to_stdout(gzip.GzipFile('', 'r', 0, BytesIO(r.content)).read())
            sys.exit(0)
            
    def delete(self):
        allres = []
        for m in self.options.models:
            if self.options.dryrun or not self.options.force:
                params = {'recursive': self.options.recursive, 'dryrun': True}
                r = requests.get(os.path.join(self.options.url,
                                              "model/delete/%s/%s/%s" % (self.options.source,
                                                                         self.options.target,
                                                                         m)),
                                 params=params, auth=self.auth)
                if r.status_code == 200:
                    mres = r.json()
                else:
                    launcher.LOGGER.error('cannot remove %s (%s)' % (m, r.text))
                    continue
                launcher.LOGGER.info('-- %sremoving %s and %d '
                                     'childrens:\n\t%s' % (self.options.dryrun and "not " or "",
                                                           m, len(mres) - 1, "\n\t".join(mres)))
            confirm = self.options.force
            if self.options.dryrun:
                continue
            confirm = confirm or launcher.confirm()
            if confirm:
                params = {'recursive': self.options.recursive}
                r = requests.get(os.path.join(self.options.url,
                                              "model/delete/%s/%s/%s" % (self.options.source,
                                                                         self.options.target,
                                                                         m)),
                                 params=params, auth=self.auth)
                if r.status_code == 200:
                    mres = r.json()
                    launcher.LOGGER.info('  => %d models removed: %s' % (len(mres), " ".join(mres)))
                    allres += mres
                else:
                    launcher.LOGGER.error('cannot remove %s (%s)' % (m, r.text))
            else:
                launcher.LOGGER.info("  ... skipping")
        res = "Total %d models removed" % len(allres)
        return res
    
    def add(self):
        if not os.path.exists(self.options.file):
            raise RuntimeError('file `%s` does not exists' % self.options.file)
        if not self.options.file.endswith(".tgz"):
            raise RuntimeError('file `%s` should be a .tgz file' % self.options.file)
        filename = os.path.basename(self.options.file)[:-4]
        parts = filename.split('_')
        if len(parts) < 4 or len(parts) > 5:
            raise RuntimeError('incorrect model naming: %s' % filename)
        trid = parts.pop(0)
        lp = parts.pop(0)
        name = parts.pop(0)
        nn = parts.pop(0)
        tar = tarfile.open(self.options.file, "r:gz")
        try:
            f = tar.extractfile("%s/config.json" % filename)
            content = f.read()
            config_json = json.loads(content)
        except Exception as e:
            raise ValueError('cannot extract `%s/config.json` from model: %s' % (filename, str(e)))
        if config_json["model"] != filename:
            raise ValueError(
                'model name does not match directory %s/%s' % (config_json["model"], filename))

        params = {
            "ignore_parent": self.options.ignore_parent,
            "compute_checksum": self.options.compute_checksum,
            "name": self.options.name
        }
        files = {'tgz': (filename, open(self.options.file, mode='rb'), 'application/octet-stream')}
        r = requests.post(os.path.join(self.options.url, "model", "add", filename),
                          auth=self.auth, params=params, files=files)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'model/add\' service: %s' % r.text)
        res = r.json()
        return res
    
    def detail(self):
        res = utils.show_model_files_list(url=self.options.url, auth=self.auth,
                                          model_name=self.options.model,
                                          directory=self.options.directory)
        return res

    def execute_command(self):
        result = None
        if self.options.subcmd == "list":
            result = self.list()
        if self.options.subcmd == "get":
            result = self.get()
        if self.options.subcmd == "delete":
            result = self.delete()
        if self.options.subcmd == "add":
            result = self.add()
        if self.options.subcmd == "describe":
            result = self.describe()
        if self.options.subcmd == "tagadd":
            result = self.tagadd()
        if self.options.subcmd == "tagdel":
            result = self.tagdel()
        if self.options.subcmd == "share":
            result = self.share()
        if self.options.subcmd == "removeshare":
            result = self.removeshare()
        if self.options.subcmd == "open":
            result = self.open()
        if self.options.subcmd == "removeopen":
            result = self.removeopen()
        if self.options.subcmd == "detail":
            result = self.detail()
        return result



import logging
import math
import os
import json
import sys
from datetime import datetime, date

import requests
import six
from jsonschema import Draft4Validator
from prettytable import PrettyTable
from packaging import version

import client.launcher as launcher

_reqcache = {}
lookup_cache = {}
test_storages = None

launcher.LOGGER = logging.getLogger()


def getcache(url, auth, params):
    global _reqcache
    key = json.dumps([url, params])
    if key not in _reqcache:
        _reqcache[key] = requests.get(url, auth=auth, params=params)
    return _reqcache[key]


def entity_list(url, auth):
    r = getcache(os.path.join(url, "entity/list"),
                 auth=auth, params={"detail": True})
    return r.status_code, r.json()


def get_entity(url, auth, code):
    try:
        code_id = int(code)
        return True, code_id
    except Exception:
        pass
    status_code, entities = entity_list(url, auth)
    if status_code != 200:
        return False, "cannot get entity list"
    for e in entities:
        if e["entity_code"] == code or e["name"] == code:
            return True, e["id"]
    return False, "unknown entity"


def group_list(url, auth):
    r = getcache(os.path.join(url, "group/list"),
                 auth=auth, params={"detail": True})
    return r.status_code, r.json()


def get_group(url, auth, code):
    try:
        code_id = int(code)
        return True, code_id
    except Exception:
        pass
    codesplit = code.split(':')
    entity_id = None
    if len(codesplit) == 2:
        status_code, entity_id = get_entity(url, auth, codesplit[0])
        code = codesplit[1]
        if not status_code:
            return False, entity_id
    elif len(codesplit) > 2:
        return False, "invalid code"

    status_code, codes = group_list(url, auth)
    if status_code != 200:
        return False, "cannot get group list"
    for c in codes:
        if c.get("entity_id") == entity_id and c["name"] == code:
            return True, c["id"]
    return False, "unknown group"


def role_list(url, auth):
    r = getcache(os.path.join(url, "role/list"),
                 auth=auth, params={"detail": True})
    return r.status_code, r.json()


def get_role(url, auth, code):
    try:
        code_id = int(code)
        return True, code_id
    except Exception:
        pass
    codesplit = code.split(':')
    entity_id = None
    if len(codesplit) == 2:
        status_code, entity_id = get_entity(url, auth, codesplit[0])
        code = codesplit[1]
        if not status_code:
            return False, entity_id
    elif len(codesplit) > 2:
        return False, "invalid code"

    status_code, codes = role_list(url, auth)
    if status_code != 200:
        return False, "cannot get role list"
    for c in codes:
        if c.get("entity_id") == entity_id and c["name"] == code:
            return True, c["id"]
    return False, "unknown role"


def user_list(url, auth):
    r = getcache(os.path.join(url, "user/list"),
                 auth=auth, params={"detail": True})
    return r.status_code, r.json()


def get_user(url, auth, user):
    try:
        user_id = int(user)
        return True, user_id
    except Exception:
        pass

    status_code, users = user_list(url, auth)
    if status_code != 200:
        return False, "cannot get user list"
    for u in users:
        # warning: .lower() works for email but not for some special characters
        # https://stackoverflow.com/questions/319426/how-do-i-do-a-case-insensitive-string-comparison
        if u["email"].lower() == user.lower() or u["tid"].lower() == user.lower():
            return True, u["id"]
    return False, "unknown user"


def permission_list(url, auth):
    r = getcache(os.path.join(url, "permission/list"),
                 auth=auth, params={"detail": True})
    return r.status_code, r.json()


def get_permission(url, auth, permission):
    try:
        permission_id = int(permission)
        return True, permission_id
    except Exception:
        pass

    status_code, permissions = permission_list(url, auth)
    if status_code != 200:
        return False, "cannot get permission list"
    for p in permissions:
        if p["name"] == permission:
            return True, p["id"]
    return False, "unknown permission"


# write data to stdout for both python 2 and 3
def write_to_stdout(data):
    if hasattr(sys.stdout, 'buffer'):
        sys.stdout.buffer.write(six.ensure_binary(data))
    else:
        sys.stdout.write(data)


def get_services(url, user_auth):
    response_services = requests.get(os.path.join(url, "service/list"),
                                     auth=user_auth, params={"minimal": True, "all": True})

    if response_services.status_code != 200:
        return False, (response_services.status_code, response_services.text)

    services = response_services.json()
    return True, services


def show_model_files_list(url, auth, model_name, directory):
    r = requests.get(os.path.join(url, "model/listfiles", model_name),
                     params={'directory': directory},
                     auth=auth)
    if r.status_code != 200:
        raise RuntimeError('incorrect result from \'model/listfiles\' service: %s' % r.text)

    res = PrettyTable(["Name", 'LastModified', 'Size (in byte)'])
    res.align["Path"] = "l"
    res.align["LastModified"] = "l"
    res.align["Size"] = "l"

    for k, v in six.iteritems(r.json()):
        file_name = k.split("/", 1)[1]
        if not file_name:
            continue
        date = datetime.fromtimestamp(v['last_modified']).strftime("%m/%d/%Y, %H:%M:%S") if v.get(
            'last_modified') else ''
        size = v.get('size') if v.get('size') else ''
        res.add_row([file_name, date, size])

    return res


def is_model_pruned(url, auth, model_name):
    r = requests.get(os.path.join(url, "model/list"),
                     params={'model': model_name},
                     auth=auth)
    if r.status_code != 200:
        raise RuntimeError('incorrect result from \'model/listfiles\' service: %s' % r.text)
    result = r.json()
    if result and result[0] and "pruned" in result[0] and result[0]["pruned"] is True:
        return True
    return False


def _get_test_storages(url, auth, service):
    global test_storages
    if test_storages is None:
        data = {'service': service}
        r = requests.get(os.path.join(url, "resource/list"), auth=auth, data=data)
        result = r.json()
        if r.status_code != 200:
            return False, (r.status_code, result.get('message'))
        test_storages = [{'name': v['name'], "entity": v["entity"]}
                         for v in result if v['type'] == "test"]
    return test_storages


def _lookup_repository(url, auth, service, remote_path, entity):
    id = (remote_path, entity)
    if id in lookup_cache:
        return True, lookup_cache[id]
    data = {'path': remote_path, 'service': service}
    if entity:
        data["entity"] = entity
    r = requests.get(os.path.join(url, "resource/list"), auth=auth, data=data)
    result = r.json()
    if r.status_code != 200:
        return False, (r.status_code, result['message'])
    lookup_cache[id] = [type(r) == dict and r.get('key') or fileObj.get('key')
                        for fileObj in result]
    return True, lookup_cache[id]


def _get_testfiles(url, auth, service, path, model, src_lang, tgt_lang):
    assert src_lang is not None and tgt_lang is not None, "src/tgt_lang not determined"
    test_storages = _get_test_storages(url, auth, service)
    if path.endswith("." + src_lang):
        (real_path, basename) = os.path.split(path)
    else:
        real_path = path
    res = []
    for t in test_storages:
        entity = t["entity"]
        storage_name = t["name"]
        status, result = _lookup_repository(url, auth, service, "%s:%s" %
                                            (storage_name, path), entity)
        if not status:
            if result[0] == 404:
                if entity == "CONF_DEFAULT":
                    entity = ""
                launcher.LOGGER.info('no test corpus found in %s (entity "%s")' %
                                     (storage_name + ":" + path, entity))
            else:
                launcher.LOGGER.error('cannot connect to test repository %s (entity "%s")' %
                                      (storage_name + ":" + path, entity))
                sys.exit(1)
        else:
            count = 0
            for f in result:
                if f.endswith("." + src_lang):
                    test_parts = f.split('/')
                    test_name = test_parts[len(test_parts) - 1]
                    res.append(
                        ("%s:%s" % (storage_name, f), "pn9_testtrans:" +
                            model + "/" + storage_name + "/" + real_path + "/" + test_name + "." + tgt_lang))
                    count += 1
            if entity == "CONF_DEFAULT":
                entity = ""
            launcher.LOGGER.info('found %d test files in %s (entity "%s")' %
                                 (count, storage_name + ":" + path, entity))
    return res


def _get_outfiles(url, auth, service, path, model, src_lang, tgt_lang):
    assert src_lang is not None and tgt_lang is not None, "src/tgt_lang not determined"
    storage = "pn9_testtrans"
    res = []
    # If input path is file, get folder
    if path.endswith("." + src_lang + "." + tgt_lang):
        (real_path, basename) = os.path.split(path)
    else:
        real_path = path
    test_storages = _get_test_storages(url, auth, service)
    # Get test storage
    output_paths = list(map(lambda t: model + "/" + t["name"] + "/" + real_path, test_storages))
    output_paths.append(model + "/" + real_path)
    # Get translated path
    for output_path in output_paths:
        status, result = _lookup_repository(url, auth, service, "%s:%s" % (storage, output_path), None)
        if not status:
            if result[0] == 404:
                launcher.LOGGER.info('no output found in %s:%s' % (storage, output_path))
            else:
                launcher.LOGGER.error("cannot connect to testtrans repository: %s" % result[1])
                sys.exit(1)
        else:
            for f in result:
                if f.endswith("." + src_lang + "." + tgt_lang):
                    res.append("pn9_testtrans:" + f)
    return res


def _get_multi_ref(reffile, result, storage):
    reffile_str = ''
    if reffile in result:
        reffile_str = storage + reffile
    else:
        return None
    idx = 1
    while True:
        reffile_temp = reffile + '.' + str(idx)
        if reffile_temp in result:
            reffile_str += ',' + storage + reffile_temp
        else:
            break
        idx += 1
    return reffile_str


def _get_reffiles(url, auth, service, src_lang, tgt_lang, list_testfiles):
    res = []
    for (testfile, outfile) in list_testfiles:
        if not testfile.endswith("." + src_lang):
            continue
        split = testfile.split(':')
        if len(split) >= 2:
            storage = split[0] + ':'
            testfile = ':'.join(split[1:])
            dirname = os.path.dirname(testfile)
            status, result = _lookup_repository(url, auth, service, storage + dirname, None)
            if status:
                reffile = testfile[:-len(src_lang)] + tgt_lang
                reffile = _get_multi_ref(reffile, result, storage)
                if reffile is not None:
                    res.append((outfile, reffile))
        else:
            # localfile
            dirname = os.path.dirname(testfile)
            result = os.listdir(dirname)
            reffile = os.path.basename(testfile[:-len(src_lang)] + tgt_lang)
            reffile = _get_multi_ref(reffile, result, dirname + '/')
            if reffile is not None:
                res.append((outfile, reffile))
    return res


def _get_datafiles(url, auth, service, path, lang):
    assert lang is not None, "lang not determined"
    res = []
    status, result = _lookup_repository(url, auth, service, path, None)
    if not status:
        if result[0] == 404:
            launcher.LOGGER.info("no data corpus found in %s" % path)
        else:
            launcher.LOGGER.error("cannot connect to data repository %s" % result[1])
            sys.exit(1)
    else:
        count = 0
        for f in result:
            if f.endswith("." + lang + ".gz"):
                filename = os.path.basename(f)
                res.append(filename)
                count += 1
        launcher.LOGGER.info('found %d files with suffix %s in %s' % (count, lang, path))
    return res


def tree_display(res, lvl, l, idx_result, model_maxsize, scorenames, bestscores,
                 skip_noscores, has_noscores, show_owner, show_pruned, only_show_pruned, quiet=False):
    sorted_l = sorted(l, key=lambda k: float(idx_result[k]["date"]))
    pref = ' ' * lvl
    for k in sorted_l:
        item = idx_result[k]
        if not skip_noscores or len(item["scores"]) != 0:
            if item["date"] is not None and item["date"] != 0:
                d = date.fromtimestamp(math.ceil(float(item["date"]))).isoformat()
            else:
                d = ""
            model = pref + item["model"]
            if "count" in item:
                model = model + " (%d)" % item["count"]
            imageTag = item["imageTag"]
            p = imageTag.find(':')
            if p != -1:
                imageTag = imageTag[:p]
            p = imageTag.rfind('/')
            if p != -1:
                imageTag = imageTag[p + 1:]
            scorecols = []
            noscores = 0
            for s in scorenames:
                score = ""
                if s in item["scores"]:
                    score = "%.02f" % float(item["scores"][s])
                    if item["scores"][s] == bestscores[s]:
                        score = '*' + score
                    elif item["scores"][s] / bestscores[s] > 0.995:
                        score = '~' + score
                else:
                    noscores += 1
                scorecols.append(score)
            if has_noscores and noscores == 0:
                continue
            sentenceCount = ''
            if 'cumSentenceCount' in item and item['cumSentenceCount'] != 0:
                sentenceCount = "%.2fM" % (item['cumSentenceCount'] / 1000000.)
            pruned = ''
            if 'pruned' in item and show_pruned:
                pruned = item["pruned"]
            if quiet:
                res.append(item["model"])
            else:
                line_data = [d, item["lp"], imageTag, model, sentenceCount]
                if show_owner:
                    owner_obj = item.get("owner")
                    line_data.append(
                        owner_obj.get('entity_code') if owner_obj and isinstance(owner_obj, dict)
                        else owner_obj)
                if show_pruned:
                    line_data.append(pruned)
                res.add_row(line_data + scorecols)
        tree_display(res, lvl + 1, item['children_models'], idx_result, model_maxsize, scorenames,
                     bestscores, skip_noscores, has_noscores, show_owner, show_pruned, only_show_pruned, quiet)


# Calculate max depth of the trees
def tree_depth(lvl, l, idx_result):
    max_level = lvl
    for k in l:
        item = idx_result[k]
        sub_level = tree_depth(lvl + 1, item['children_models'], idx_result)
        if sub_level > max_level:
            max_level = sub_level
    return max_level


# Calculate cumulated sentenceCount
def cum_sentenceCount(l, idx_result, sentenceCount):
    for k in l:
        item = idx_result[k]
        if 'cumSentenceCount' not in item or item['cumSentenceCount'] is None:
            item['cumSentenceCount'] = item['sentenceCount'] + sentenceCount
        sub_level = cum_sentenceCount(item['children_models'], idx_result, item['cumSentenceCount'])


# Merge two configs (redundant code with method in nmt-wizard/server/nmtwizard/config.py and
# nmt-wizard-docker/nmtwizard/utils.py)
def merge_config(a, b):
    """Merges config b in a."""
    if isinstance(a, dict):
        for k, v in six.iteritems(b):
            if k in a and isinstance(v, dict) and type(a[k]) == type(v):
                merge_config(a[k], v)
            else:
                a[k] = v
    return a


def _get_params(lparam, listcmd):
    res = []
    idx = 0
    while idx < len(listcmd):
        if listcmd[idx] in lparam:
            idx = idx + 1
            while idx < len(listcmd) and not listcmd[idx].startswith('-'):
                res.append(listcmd[idx])
                idx += 1
            continue
        idx += 1
    return res


def _get_params_except_specified(lparam, listcmd):
    res = []
    idx = 0
    while idx < len(listcmd):
        if listcmd[idx] in lparam:
            idx = idx + 1
            while idx < len(listcmd) and not listcmd[idx].startswith('-'):
                idx += 1
        else:
            res.append(listcmd[idx])
            idx = idx + 1
            while idx < len(listcmd) and not listcmd[idx].startswith('-'):
                res.append(listcmd[idx])
                idx += 1
    return res


# Parse docker image name to get version pattern (e.g. "systran/pn9_tf:v1") and number (1)
def parse_version_number(image):
    p = image.find(".")
    if p == -1:
        # version incompletely qualified
        current_version_pattern = image
    else:
        # version completely qualified
        current_version_pattern = image[:p]
    q = current_version_pattern.find("v")
    if q == -1:
        version_main_number = 0
    else:
        version_main_number = int(current_version_pattern[q + 1:])
    return (current_version_pattern, version_main_number)


# Check upgrades for docker image and return upgraded version if available and accepted by user
# input image and tag
# output image:tag
def check_upgrades(auth, url, upgrade, image, tag):
    image_dec = image.split("/")
    image = '/'.join(image_dec[-2:])
    tag_prefix = ""
    if tag[0] == 'v':
        tag_prefix = "v"
        tag = tag[1:]
    try:
        version_parts = version.parse(tag).release
    except ValueError as err:
        raise RuntimeError('cannot parse version %s - %s' % (tag, str(err)))

    if version_parts and version_parts[0] >= 1 and (len(version_parts) < 3 or upgrade != "none"):
        tag_req = tag
        if len(version_parts) == 3:
            tag_req = "%d.%d" % (version_parts[0], version_parts[1])
        r = requests.get(os.path.join(url, "docker/versions"),
                         auth=auth, params={'version_pattern': image + ':' + tag_prefix + tag_req})
        result = r.json()
        if r.status_code != 200:
            raise RuntimeError('cannot retrieve docker images for current version %s -- %s' %
                               (image + ':' + tag_prefix + tag, r.text))
        if len(result) == 0:
            raise RuntimeError('unknown version %s' % (image + ':' + tag_prefix + tag))
        versions = [version.parse(r['image']) for r in result]
        latest_version_parse = max(versions)
        latest_version = latest_version_parse.base_version
        # selectively upgrade if later version available
        if version.parse(image + ':' + tag_prefix + tag) < latest_version_parse:
            if len(version_parts) < 3 or upgrade == "force":
                # version incompletely qualified
                launcher.LOGGER.info('automatically upgrading docker_image=%s to %s' %
                                     (image, latest_version))
                return latest_version, True
            else:
                # version completely qualified
                launcher.LOGGER.info('upgrading docker_image=%s to %s is available, '
                                     'do you want to upgrade? (y/n)' %
                                     (image + ':' + tag_prefix + tag, latest_version))
                while True:
                    response = input('Upgrade? ')
                    if response in {'y', 'yes'}:
                        launcher.LOGGER.info('upgrading docker_image=%s to %s' %
                                             (image + ':' + tag_prefix + tag, latest_version))
                        return latest_version, True
                    elif response in {'n', 'no'}:
                        break
                    else:
                        launcher.LOGGER.info('Please enter `y` or `n`.')
        elif version.parse(image + ':' + tag_prefix + tag) == latest_version_parse:
            if len(version_parts) < 3:
                # version incompletely qualified
                launcher.LOGGER.info('automatically upgrading docker_image=%s to %s' %
                                     (image, latest_version))
                return latest_version, True
    return image + ':' + tag_prefix + tag, False


# Announce the usage of a docker image
def announce_usage(image):
    split = image.split("/")
    if len(split) > 2:
        image = "/".join(split[-2:])
    launcher.LOGGER.info('** will be using -docker_image=%s' % image)


# Return a string with the list of all schema validation warnings
def get_schema_errors(schema, config):
    v = Draft4Validator(schema)
    all_errors = "\n\n**Your config has the following issues, please refer to the documentation:"
    for error in sorted(v.iter_errors(config), key=str):
        # format error message
        error_parts = error.message.split()
        error_message = ""
        for error_part in error_parts:
            if error_part.startswith("u'"):
                error_message += error_part[1:].replace("'", '"')
            elif error_part[1:].startswith("u'"):
                error_message += error_part[0] + error_part[2:].replace("'", '"')
            else:
                error_message += error_part
            error_message += ' '
        # format error path
        error_path = ""
        for path_part in list(error.path):
            if isinstance(path_part, int):
                error_path += "array[" + str(path_part) + "]"
            else:
                error_path += '"' + path_part + '"'
            error_path += '/'
        # write error message and path
        if error_path == "":
            all_errors += '\n - In the config, ' + error_message
        else:
            all_errors += '\n - In the option ' + error_path + ', ' + error_message
    return all_errors



import json
import logging
import os
import random
import re
import sys

import requests
from jsonschema import validate
from jsonschema.exceptions import ValidationError

import client.launcher as launcher
from lib import utils

launcher.LOGGER = logging.getLogger()


class Task:
    def __init__(self, options, auth):
        self.options = options
        self.auth = auth
        self.is_json = self.options.display == "JSON"
        self.skip_launch = False

    def change(self):
        if self.options.prefix is None and len(self.options.task_ids) == 0 \
                and self.options.gpus is None:
            raise RuntimeError('you need to specify either `--prefix PREFIX` '
                               'or task_id(s) or `--gpus NGPUS`')
        if self.options.prefix is not None and len(self.options.task_ids) != 0:
            raise RuntimeError('you cannot to specify both `--prefix PREFIX` '
                               'and task_id(s)')
        if self.options.service is None and self.options.priority is None:
            raise RuntimeError('you need to specify new service (`--service SERVICE`)'
                               ' and/or new priority (`--priority PRIORITY`)')
        if self.options.prefix:
            r = requests.get(os.path.join(self.options.url, "task/list",
                                          self.options.prefix + '*'), auth=self.auth)
            if r.status_code != 200:
                raise RuntimeError('incorrect result from \'task/list\' service: %s' % r.text)
            result = r.json()
            self.options.task_ids = [k["task_id"] for k in result]
            if len(result) == 0:
                raise RuntimeError('no task matching prefix %s' % self.options.prefix)
        launcher.LOGGER.info(
            'Change %d tasks (%s)' % (len(self.options.task_ids), ", ".join(self.options.task_ids)))
        if len(self.options.task_ids) == 1 or launcher.confirm():
            modification = ""
            if self.options.service:
                modification += "service=%s" % self.options.service
            if self.options.priority:
                if len(modification) > 0:
                    modification += ", "
                modification += "priority=%d" % self.options.priority
            if self.options.gpus:
                if len(modification) > 0:
                    modification += ", "
                modification += "ngpus=%d" % self.options.gpus
            launcher.LOGGER.info("modifying tasks (%s) for:" % modification)
            error = False
            for k in self.options.task_ids:
                launcher.LOGGER.info("*** %s" % k)
                p = self.options.priority
                if p is not None and self.options.priority_rand != 0:
                    p += random.randint(0, self.options.priority_rand)
                params = {'priority': p, 'service': self.options.service,
                          'ngpus': self.options.gpus}
                r = requests.get(os.path.join(self.options.url, "task/change", k),
                                 auth=self.auth, params=params)
                if r.status_code != 200:
                    launcher.LOGGER.error('>> %s' % r.json()["message"])
                    error = True
                else:
                    launcher.LOGGER.info(">> %s" % r.json()["message"])
                res = ""
        else:
            res = ""
        return res

    def file(self):
        p = self.options.filename.find(':')
        if p == -1:
            r = requests.get(os.path.join(self.options.url, "task/file",
                                          self.options.task_id, self.options.filename),
                             auth=self.auth)
        else:
            r = requests.get(os.path.join(self.options.url, "task/file_storage",
                                          self.options.filename[0:p], self.options.task_id,
                                          self.options.filename[p + 1:]),
                             auth=self.auth)
        if r.status_code != 200:
            raise RuntimeError('incorrect result from \'task/file_extended\' service: %s' % r.text)
        res = r.content
        return res

    def launch(self):
        res = None
        mode = None
        model = None
        src_lang = None
        tgt_lang = None
        totranslate = None

        # first pass to get model if present and if no image given, determine
        # docker image to be used
        i = 0
        while i < len(self.options.docker_command):
            tok = self.options.docker_command[i]
            if mode is None and (tok == "-m" or tok == "--model"):
                assert i + 1 < len(self.options.docker_command), "`-m` missing value"
                model = self.options.docker_command[i + 1]
                is_pruned = utils.is_model_pruned(self.options.url, self.auth, model)
                assert is_pruned is False, "cannot launch task on pruned model"

                if self.options.docker_image is None:
                    # no image specified with -i, first try to infer it from model
                    r = requests.get(os.path.join(self.options.url, "model/describe", model),
                                     params={"short": True}, auth=self.auth)
                    if r.status_code != 200:
                        raise RuntimeError("cannot infer docker_image for "
                                           "model %s -- %s" % (model, r.text))
                    self.options.docker_image = r.json()['imageTag']
                    # if docker_tag option (-t) specified, it overrule on docker image modifier
                    if self.options.docker_tag:
                        p = self.options.docker_image.find(':')
                        if p != -1:
                            self.options.docker_image = self.options.docker_image[:p]

                split = model.split("_")
                if len(split) > 2 and src_lang is None:
                    lp = split[1]
                    m = re.match(r"^([a-z]{2}([-+][A-Z]+)?)([a-z]{2}.*)$", lp)
                    if m is not None:
                        src_lang = m.group(1)
                        tgt_lang = m.group(3)
                i += 1
            i += 1
        # second pass to apply other options -c, -T, -Tm, -t
        i = 0
        config = None
        while i < len(self.options.docker_command):
            tok = self.options.docker_command[i]
            if mode is None and (tok == "train" or tok == "trans" or
                                 tok == "preprocess" or tok == "release"):
                mode = tok
                assert mode != "trans" or model is not None, "missing model for `trans`"
            # get config if present to validate it against schema
            elif mode is None and (tok == "-c" or tok == "--config"):
                assert i + 1 < len(self.options.docker_command), "`-c` missing value"

                # get JSON config passed as parameter
                c = self.options.docker_command[i + 1]
                if c.startswith("@"):
                    with open(c[1:], "rt") as f:
                        c = f.read()
                config = json.loads(c)
                if "source" in config:
                    src_lang = config["source"]
                if "target" in config:
                    tgt_lang = config["target"]
                i += 1
            elif mode == "trans" and (tok == "-T" or tok == "-Tm" or tok == "-t"):
                files = []
                if tok == '-t':
                    assert i + 1 < len(self.options.docker_command), "`trans -t` missing value"
                    input_files = self.options.docker_command[i + 1:]
                    for filename in input_files:
                        # check filename
                        p = filename.rfind(".")
                        assert p != -1, "language suffix should end filename %s" % filename
                        if src_lang is None:
                            src_lang = filename[p + 1:]
                        else:
                            assert src_lang == filename[p + 1:], \
                                "incompatible language suffix in filename %s" % filename
                        if tgt_lang is None:
                            p = model.find("_")
                            q = model.find("_", p + 1)
                            assert p != -1 and q != -1, \
                                "cannot find language pair in model name %s" % model
                            lp = model[p + 1:q]
                            assert lp[
                                   :len(src_lang)] == src_lang, "model lp (%s) does not " \
                                                                "match language suffix (%s)" % (
                                                                    lp[:len(src_lang)], src_lang)
                            tgt_lang = lp[len(src_lang):]
                        # find file to translate
                        file_path = utils._get_testfiles(self.options.url, self.auth,
                                                         self.options.service, filename,
                                                         model, src_lang, tgt_lang)
                        assert len(
                            file_path) != 0, "no file corresponds to filename %s" % filename
                        files.extend(file_path)
                else:
                    assert i + 1 < len(self.options.docker_command), "`trans -T[m]` missing value"
                    path = self.options.docker_command[i + 1]
                    assert i + 2 == len(
                        self.options.docker_command), "`trans -T[m] [PATH]` has extra values"
                    p = path.rfind(".")
                    assert p == -1, "-T[m] PATH should be a folder not a file %s" % path
                    if path[-1:] == "/":
                        path = path[:-1]
                    # find files to translate
                    files = utils._get_testfiles(self.options.url, self.auth,
                                                 self.options.service, path,
                                                 model, src_lang, tgt_lang)
                    assert len(files) != 0, "no file found to translate in folder %s" % path
                    if (tok == '-Tm'):
                        # find translated files
                        translatedfiles = utils._get_outfiles(self.options.url, self.auth,
                                                              self.options.service,
                                                              path, model, src_lang, tgt_lang)
                        files = [(test, out) for (test, out) in files if self.is_untranslated_files(out, translatedfiles)]
                        assert len(
                            files) != 0, "-Tm translates only untranslated files but " \
                                         "all files already translated in %s" % path

                # if needed prepare translation
                launcher.LOGGER.info('found %d test files' % len(files))
                self.options.toscore = utils._get_reffiles(self.options.url, self.auth,
                                                           self.options.service,
                                                           src_lang, tgt_lang, files)
                launcher.LOGGER.info('found %d test references' % len(self.options.toscore))
                docker_command = self.options.docker_command
                res = []
                input_files = []
                output_files = []
                for f in files:
                    launcher.LOGGER.info("translating: " + f[0])
                    input_files.append(f[0])
                    output_files.append(f[1])
                new_params = ["-i"] + input_files + ["-o"] + output_files
                self.options.docker_command = docker_command[0:i]
                self.options.docker_command += new_params

                self.skip_launch = True
                break
            elif mode == "trans" and (tok == "-i" or tok == "-I") and (
                    "-bt" in self.options.docker_command):
                # if "-I" is set, search files with the provided prefix
                #    "-i shared_data:fr/train/mono_fr-XX_News__28779-STATMT-newscrawl-2016_2.fr.gz"
                #    "-I shared_data:fr/train/mono_fr-XX_News__28779-STATMT-newscrawl-2016"
                #    "-I shared_data:fr/train/mono_fr"
                def generate_decoder_options(c):
                    if c is None:
                        return "default"
                    options_str = ''
                    params = c.get("options", {}).get("config", {}).get("params", {})
                    if 'beam_width' in params:
                        options_str = options_str + 'beam_' + str(params['beam_width']) + '_'
                    if 'sampling_topk' in params:
                        options_str = options_str + 'sampling_' + str(
                            params['sampling_topk']) + '_'
                    params = c.get("postprocess", {})
                    if 'remove_placeholders' in params:
                        options_str = options_str + 'remove_' + str(
                            params['remove_placeholders']) + '_'
                    if options_str != '':
                        return options_str[:-1]
                    return "default"

                new_params = utils._get_params_except_specified(("-i", "--input", "-I", "-bt"),
                                                                self.options.docker_command[i:])
                input_files_opts = utils._get_params(("-i", "--input", "-I"),
                                                     self.options.docker_command[i:])
                output_files_opts = utils._get_params("-bt", self.options.docker_command[i:])
                if len(input_files_opts) != len(output_files_opts):
                    launcher.LOGGER.error("invalid trans command - misaligned input/bt_output")
                    sys.exit(1)

                if tok == "-I":
                    assert src_lang is not None, "cannot find source lang id"
                else:
                    filename = input_files_opts[0]
                    segments = filename.split(".")
                    assert len(segments) >= 3, "input file should have suffix like .fr.gz"
                    assert segments[-1] == "gz", "input file should be gzipped"
                    if src_lang is None:
                        src_lang = segments[-2]
                    else:
                        assert src_lang == segments[-2], "incompatible language suffix"
                if tgt_lang is None:
                    segments = model.split("_")
                    assert len(segments) == 5, "illegal model name"
                    lp = segments[1]
                    assert lp[:len(src_lang)] == src_lang, "model lp does not match " \
                                                           "language suffix"
                    tgt_lang = lp[len(src_lang):]

                if "--copy_source" not in self.options.docker_command:
                    new_params += ["--copy_source"]
                self.options.docker_command = self.options.docker_command[0:i]

                if "--keep_placeholders_as_config" in self.options.docker_command:
                    self.options.docker_command.remove('--keep_placeholders_as_config')
                elif "--keep_placeholders_as_config" in new_params:
                    new_params.remove('--keep_placeholders_as_config')
                else:
                    if config is None:
                        config = {}
                        self.options.docker_command.insert(0, "-c")
                        self.options.docker_command.insert(1, "{}")
                    if "postprocess" in config:
                        config['postprocess']['remove_placeholders'] = True
                    else:
                        config['postprocess'] = {'remove_placeholders': True}
                    option_index = self.options.docker_command.index("-c") + 1
                    self.options.docker_command[option_index] = json.dumps(config)

                decoder_options = generate_decoder_options(config)
                if "--add_bt_tag" in self.options.docker_command or "--add_bt_tag" in new_params:
                    decoder_options = decoder_options + "_tagged"

                input_files = []
                output_files = []
                for f, b in zip(input_files_opts, output_files_opts):
                    files = []
                    if tok == "-I":
                        path, filename = os.path.split(f)
                        launcher.LOGGER.info("searching files in " + path)
                        monofiles = utils._get_datafiles(self.options.url, self.auth,
                                                         self.options.service,
                                                         path, src_lang)
                        files = [os.path.join(path, test) for test in monofiles if
                                 test.startswith(filename)]
                    else:
                        launcher.LOGGER.info("translating: " + f)
                        files.append(f)

                    output_path_name = os.path.join(b, src_lang + tgt_lang, model,
                                                    decoder_options)
                    transfiles = utils._get_datafiles(self.options.url, self.auth,
                                                      self.options.service,
                                                      output_path_name, tgt_lang)

                    for input in files:
                        filename = os.path.basename(input)
                        filename, _ = os.path.splitext(filename)  # remove ".gz"
                        filename, _ = os.path.splitext(filename)  # remove ".fr"
                        output_file_name = filename + "." + tgt_lang + ".gz"

                        if output_file_name in transfiles:
                            launcher.LOGGER.info("%s exists in %s, skip translating..." % (
                                output_file_name, output_path_name))
                        else:
                            input_files.append(input)
                            output_files.append(
                                os.path.join(output_path_name, output_file_name))
                if not input_files:
                    launcher.LOGGER.info("all files have already been there! exit!")
                    sys.exit(0)
                new_params += ["-i"] + input_files + ["-o"] + output_files
                self.options.docker_command += new_params

                break
            i += 1

        # Docker image checks
        if self.options.docker_image is None:
            raise RuntimeError('missing docker image (you can set LAUNCHER_IMAGE)')
        # if docker_tag option (-t) specified, it overrule on docker image modifier
        p = self.options.docker_image.find(':')
        if p != -1:
            if self.options.docker_tag is not None:
                raise RuntimeError("ambiguous definition of docker tag (-i %s/-t %s)",
                                   (self.options.docker_image, self.options.docker_tag))
            self.options.docker_tag = self.options.docker_image[p + 1:]
            self.options.docker_image = self.options.docker_image[:p]

        # check if we can upgrade version
        self.options.docker_image, upgraded = utils.check_upgrades(self.auth, self.options.url,
                                                                   self.options.upgrade,
                                                                   self.options.docker_image,
                                                                   self.options.docker_tag)
        self.options.docker_tag = None
        utils.announce_usage(self.options.docker_image)

        # if we are translating check if there are reference files
        if mode == "trans":
            idx = self.options.docker_command.index("trans")
            input_files = utils._get_params(("-i", "--input"),
                                            self.options.docker_command[idx + 1:])
            output_files = utils._get_params(("-o", "--output"),
                                             self.options.docker_command[idx + 1:])
            if len(input_files) != len(output_files):
                launcher.LOGGER.error("invalid trans command - misaligned input/output")
                sys.exit(1)
            self.options.toscore = utils._get_reffiles(self.options.url, self.auth,
                                                       self.options.service,
                                                       src_lang, tgt_lang,
                                                       list(zip(input_files, output_files)))
            launcher.LOGGER.info('found %d test references' % len(self.options.toscore))
            if "--chaintuminer" in self.options.docker_command:
                self.options.totuminer = list(zip(input_files, output_files))
                self.options.docker_command.remove("--chaintuminer")

        if mode == "trans" and (tok == "-T" or tok == "-Tm" or tok == "-t"):
            status, serviceList = utils.get_services(url=self.options.url, user_auth=self.auth)
            res.append(launcher.process_request(serviceList, self.options.cmd, self.options.subcmd,
                                                self.options.display == "JSON",
                                                self.options, auth=self.auth))

        # merge -c and -m configs and validate it against schema
        if self.options.novalidschema:
            launcher.LOGGER.warning(
                "schema validation is skipped, your config is potentially erroneous")
        else:
            if not (config is None) or upgraded:
                if model:
                    # if model is present, collect its config
                    r = requests.get(os.path.join(self.options.url, "model/describe", model),
                                     auth=self.auth)
                    if r.status_code != 200:
                        raise RuntimeError("cannot retrieve configuraton for "
                                           "model %s -- %s" % (model, r.text))
                    model_config = r.json()
                    if config:
                        # merge to validate complete config
                        config = utils.merge_config(model_config, config)
                    else:
                        config = model_config

                image = self.options.docker_image
                _, version_main_number = utils.parse_version_number(image)
                if version_main_number > 0:
                    p = image.find(":")
                    tag = image[p + 1:]
                    image = image[:p]
                    r = requests.get(os.path.join(self.options.url, "docker/schema"),
                                     params={'image': image, 'tag': tag}, auth=self.auth)
                    if r.status_code != 200:
                        raise RuntimeError('cannot retrieve schema from docker image %s,'
                                           ' tag %s: %s' % (image, tag))
                    schema_res = r.json()
                    schema = json.loads(schema_res)
                    # validate config against JSON schema
                    try:
                        validate(config, schema)
                    except ValidationError as error:
                        all_errors = utils.get_schema_errors(schema, config)
                        raise ValidationError(all_errors)

        if mode == "release":
            self.options.docker_command += ["-d", "pn9_release:"]
        if self.options.no_test_trans:
            assert mode == "train", "`--no_test_trans` can only be used with `train` mode"
        elif mode == "train":
            assert not (src_lang is None or tgt_lang is None), "src/tgt_lang not determined: " \
                                                               "cannot find test sets"
            if src_lang < tgt_lang:
                test_dir = src_lang + "_" + tgt_lang
            else:
                test_dir = tgt_lang + "_" + src_lang
            self.options.totranslate = utils._get_testfiles(self.options.url, self.auth,
                                                            self.options.service, test_dir,
                                                            "<MODEL>", src_lang, tgt_lang)
            self.options.toscore = utils._get_reffiles(self.options.url, self.auth,
                                                       self.options.service, src_lang,
                                                       tgt_lang, self.options.totranslate)
            launcher.LOGGER.info('found %d test references' % len(self.options.toscore))
        return res


    def is_untranslated_files(self, out, translated_paths):
        if len(translated_paths) == 0:
            return True
        # Check output is new format and existed
        if (out in translated_paths):
            return False
        # Check output is old format and existed: check by name and number of character /
        # the number of character / in old format is less than new format
        # (Eg: old format 'pn9_testtrans:model/path/test_name.en.de', new format 'pn9_testtrans:model/storage/path/test_name.en.de')
        translated_files = list(filter(lambda path: self.compare_test_name(out, path) and out.count('/') != path.count('/'), translated_paths))

        return len(translated_files) == 0


    def compare_test_name(self, out, translated_path):
        out_name = os.path.basename(out)
        translated_path_name = os.path.basename(translated_path)

        return out_name == translated_path_name


    def execute_command(self):
        result = None
        if self.options.subcmd == "change":
            result = self.change()
        if self.options.subcmd == "file":
            result = self.file()
        if self.options.subcmd == "launch":
            result = self.launch()
        return result
